@@ -1,15 +1,16 @@
 import abc
 import json
 import os
+import ast
 from datetime import datetime
 from pathlib import Path
 import pydantic
-from typing import Union
 
-from flowui.client.airflow_client import AirflowRestClient
 from flowui.logger import get_configured_logger
 from flowui.utils.enum_types import DeployModeType
 from flowui.exceptions.exceptions import InvalidOperatorOutputError
+# from flowui.client.fs_client import FileSystemClient
+import time
 
 
 class BaseOperator(metaclass=abc.ABCMeta):
@@ -25,12 +26,6 @@ class BaseOperator(metaclass=abc.ABCMeta):
         # Operator name as used by Airflow
         cls.__name__ = metadata.get("name", "BaseOperator")
 
-        # Operator version: update it to auto update the deployed version
-        cls.__version__ = metadata.get("version", "0.1.0")
-
-        # Dockerfile to build Image for this function
-        cls.__dockerfile__ = metadata.get("dockerfile", "Dockerfile-base")
-
         # Full metadata
         cls._metadata_ = metadata
 
@@ -43,10 +38,7 @@ class BaseOperator(metaclass=abc.ABCMeta):
     ) -> None:
         """
         The base class from which every FlowUI custom Operator should inherit from.
-        BaseOperator methods and variables that can be used by inheriting Operators:
-
-        self._metadata_ - Metadata as defined by the user
-        self.upstream_tasks - Metadata and XCom data results from previous Tasks
+        BaseOperator methods and variables that can be used by inheriting Operators classes:
         self.results_path - Path to store results data
         self.logger - Logger functionality
 
@@ -61,9 +53,6 @@ class BaseOperator(metaclass=abc.ABCMeta):
         self.dag_id = dag_id
         self.deploy_mode = deploy_mode
 
-        # Clients
-        self.airflow_client = AirflowRestClient()
-
         # Logger
         self.logger = get_configured_logger(f"{self.__class__.__name__ }-{self.task_id}")
 
@@ -74,43 +63,62 @@ class BaseOperator(metaclass=abc.ABCMeta):
         """
         self.logger.info(f"Started {self.task_id} of type {self.__class__.__name__} at {str(datetime.now().isoformat())}")
 
+    def _wait_for_sidecar_paths(self):
+        # Wait for sidecar create directories
+        while True:
+            if Path(self.report_path).is_dir():
+                break
+            time.sleep(2)
 
     def generate_paths(self):
         """
-        Generates paths and stores them in the attributes `self.volume_mount_path`, `self.run_path`, `self.results_path` and `self.report_path`.
+        Generates paths for shared storage.
         """
-        self.volume_mount_path = os.environ.get("VOLUME_MOUNT_PATH_DOCKER", "/opt/mnt/fs")
-        
-        self.run_path = f"{self.volume_mount_path}/runs/{self.dag_id}/{self.dag_run_id}"
-        if not Path(self.run_path).is_dir():
-            Path(self.run_path).mkdir(parents=True, exist_ok=True)
-        
-        self.results_path = f"{self.run_path}/{self.task_id}/results"
+        # Base path for fetching and storing runs results
+        if not Path(self.workflow_shared_storage).is_dir():
+            Path(self.workflow_shared_storage).mkdir(parents=True, exist_ok=True)
+
+        # # Path to store results data
         if not Path(self.results_path).is_dir():
             Path(self.results_path).mkdir(parents=True, exist_ok=True)
-        
-        self.report_path = f"{self.run_path}/{self.task_id}/report"
+
+        # # Path to store XCOM data - XCOM data is saved here AND also at /airflow/xcom/return.json
+        if not Path(self.xcom_path).is_dir():
+            Path(self.xcom_path).mkdir(parents=True, exist_ok=True)
+
+        # # Path to store report data
         if not Path(self.report_path).is_dir():
             Path(self.report_path).mkdir(parents=True, exist_ok=True)
 
 
-    def get_upstream_tasks_xcom(self):
+    def get_upstream_tasks_data(self):
         """
-        Get XCOM from upstream tasks. Stores this information in the attribute `self.upstream_task_xcom`
+        Get XCOM and results data from upstream tasks. Stores this information in the attribute `self.upstream_task_data` as a dictionary with the following structure:
+        {
+            "task_id_1": {
+                "xcom": {...},
+                "results": "path/to/results"
+            },
+            "task_id_2": {...},
+        }
 
         Raises:
             NotImplementedError: _description_
         """
-        self.upstream_task_xcom = dict()
-        if self.deploy_mode == "local-python":
-            for tid in list(self.airflow_context['task'].upstream_task_ids):
-                self.upstream_task_xcom[tid] = self.airflow_context['ti'].xcom_pull(task_ids=tid) 
-        elif self.deploy_mode == "local-bash":
-            with open("/opt/airflow/flowui/xcom/xcom_input.json") as f:
-                self.upstream_task_xcom = json.load(f)
-        elif self.deploy_mode == "local-k8s":
-            with open("/opt/airflow/flowui/xcom/xcom_input.json") as f:
-                self.upstream_task_xcom = json.load(f)
+        self.upstream_tasks_data = dict()
+        # if self.deploy_mode == "local-python":
+        #     for tid in list(self.airflow_context['task'].upstream_task_ids):
+        #         self.upstream_tasks_xcom[tid] = self.airflow_context['ti'].xcom_pull(task_ids=tid) 
+        # elif self.deploy_mode == "local-bash":
+        #     with open("/opt/airflow/flowui/xcom/xcom_input.json") as f:
+        #         self.upstream_tasks_xcom = json.load(f)
+        if self.deploy_mode == "k8s":
+            upstream_tasks_ids = ast.literal_eval(os.getenv("AIRFLOW_UPSTREAM_TASKS_IDS", str(list())))
+            for tid in upstream_tasks_ids:
+                self.upstream_tasks_data[tid] = dict()
+                self.upstream_tasks_data[tid]["results"] = self.workflow_shared_storage + f"/{tid}/results"
+                with open(f"{self.workflow_shared_storage}/{tid}/xcom/return.json") as f:
+                    self.upstream_tasks_data[tid]["xcom"] = json.load(f)
         else:
             raise NotImplementedError(f"Get upstream XCOM not implemented for deploy_mode=={self.deploy_mode}")
 
@@ -118,6 +126,10 @@ class BaseOperator(metaclass=abc.ABCMeta):
     def validate_and_get_env_secrets(self, operator_secrets_model: pydantic.BaseModel = None):
         """
         Get secret variables for this Operator from ENV. The necessary secret variables to run the Operator should be defined in the Operator's SecretsModel.
+        The secrets can then be retrieved and used in the Operator's `operator_function` method as such:
+        ```
+        my_secret = self.secrets.my_secret
+        ```
 
         Args:
             operator_secrets_model (pydantic.BaseModel): _description_
@@ -126,30 +138,50 @@ class BaseOperator(metaclass=abc.ABCMeta):
         if operator_secrets_model:
             secrets_names = list(operator_secrets_model.schema()["properties"].keys())
             secrets = dict()
+            secrets_values = ast.literal_eval(os.environ.get('FLOWUI_K8S_OPERATOR_SECRETS'))
+            if not secrets_values:
+                secrets_values = {}
             for s in secrets_names:
-                secrets[s] = os.getenv(s, None)
+                secrets[s] = secrets_values.get(s, None)
             self.secrets = operator_secrets_model(**secrets)
 
 
-    def format_xcom(self, xcom_obj: dict) -> dict:
+    def format_xcom(self, output_obj: pydantic.BaseModel) -> dict:
         """
-        Adds extra metadata to XCOM dictionary content.
+        Formats and adds extra metadata to XCOM dictionary content.
 
         Args:
-            xcom_obj (dict): XCOM dictionary
+            output_obj (dict): Pydantic output model
 
         Returns:
             dict: XCOM dictionary
         """
+        xcom_obj = output_obj.dict()
         if not isinstance(xcom_obj, dict):
             print(f"Operator {self.__class__.__name__} is not returning a valid XCOM object. Auto-generating a base XCOM for it...")
             self.logger.info(f"Operator {self.__class__.__name__} is not returning a valid XCOM object. Auto-generating a base XCOM for it...")
             xcom_obj = dict()
 
+        # Add arguments types to XCOM 
+        # TODO - this is a temporary solution. We should find a better way to do this
+        output_schema = output_obj.schema()
+        for k, v in output_schema["properties"].items():
+            if "type" in v:
+                # Get file-path and directory-path types
+                if v["type"] == "string" and "format" in v:
+                    v_type = v["format"]
+                else:
+                    v_type = v["type"]
+            elif "anyOf" in v:
+                if "$ref" in v["anyOf"][0]:
+                    type_model = v["anyOf"][0]["$ref"].split("/")[-1]
+                    v_type = output_schema["definitions"][type_model]["type"]
+            xcom_obj[f"{k}_type"] = v_type
+
+        # Update XCOM with extra metadata
         xcom_obj.update(
             operator_name=self.__class__.__name__,
-            operator_metadata=self._metadata_,
-            results_path=self.results_path,
+            operator_metadata=self._metadata_
         )
         return xcom_obj
 
@@ -172,13 +204,18 @@ class BaseOperator(metaclass=abc.ABCMeta):
             file_path.parent.mkdir(parents=True, exist_ok=True)
             with open(str(file_path), 'w') as fp:
                 json.dump(xcom_obj, fp, indent=4)
-        elif self.deploy_mode == "kubernetes":
+        elif self.deploy_mode == "k8s":
             # In Kubernetes, return XCom must be stored in /airflow/xcom/return.json
             # https://airflow.apache.org/docs/apache-airflow-providers-cncf-kubernetes/stable/operators.html#how-does-xcom-work
             file_path = Path('/airflow/xcom/return.json')
             file_path.parent.mkdir(parents=True, exist_ok=True)
             with open(str(file_path), 'w') as fp:
                 json.dump(xcom_obj, fp)
+
+            # Also store it at /home/workflow_shared_data/{self.task_id}/xcom/return.json for convenience
+            file_path = self.xcom_path + "/return.json"
+            with open(file_path, 'w') as fp:
+                json.dump(xcom_obj, fp)
         else:
             raise NotImplementedError("deploy mode not accepted for xcom push")
     
@@ -189,7 +226,7 @@ class BaseOperator(metaclass=abc.ABCMeta):
         op_kwargs: dict,
         operator_input_model: pydantic.BaseModel,
         operator_output_model: pydantic.BaseModel, 
-        operator_secrets_model: pydantic.BaseModel = None,  
+        operator_secrets_model: pydantic.BaseModel = None
     ):
         """
         _summary_
@@ -207,6 +244,10 @@ class BaseOperator(metaclass=abc.ABCMeta):
         # Start logger
         self.start_logger()
 
+        self.operator_input_model = operator_input_model
+        self.operator_output_model = operator_output_model
+        self.operator_secrets_model = operator_secrets_model
+
         # Airflow context dictionary: https://composed.blog/airflow/execute-context
         # For local-bash and kubernetes deploy modes, we assemble this ourselves and the context data is more limited
         self.airflow_context = airflow_context
@@ -216,35 +257,55 @@ class BaseOperator(metaclass=abc.ABCMeta):
         self.validate_and_get_env_secrets(operator_secrets_model=operator_secrets_model)
 
         # Generate paths
-        self.generate_paths()
-
-        # Get XCOM from upstream tasks
-        self.get_upstream_tasks_xcom()
+        self.workflow_shared_storage = "/home/shared_storage"
+        self.results_path = f"{self.workflow_shared_storage}/{self.task_id}/results"
+        self.xcom_path = f"{self.workflow_shared_storage}/{self.task_id}/xcom"
+        self.report_path = f"{self.workflow_shared_storage}/{self.task_id}/report"
+        shared_storage_source = os.environ.get('FLOWUI_WORKFLOW_SHARED_STORAGE', None)
+        if not shared_storage_source or shared_storage_source == "none":
+            self.generate_paths()
+        else:
+            self._wait_for_sidecar_paths()
 
         # Using pydantic to validate input data
-        # operator_model_obj = operator_input_model(**op_kwargs)
-        updated_op_kwargs = dict()
-        for k, v in op_kwargs.items():
-            if isinstance(v, dict) and v.get("type", None) == "FromUpstream":
-                upstream_task_id = v.get("upstream_task_id")
-                output_arg = v.get("output_arg")
-                updated_op_kwargs[k] = self.upstream_task_xcom[upstream_task_id][output_arg]
-            else:
-                updated_op_kwargs[k] = v
-        input_model_obj = operator_input_model(**updated_op_kwargs)
+        input_model_obj = operator_input_model(**op_kwargs)
 
         # Run operator function
-        output_model_obj = self.operator_function(input_model=input_model_obj)
+        output_obj = self.operator_function(input_model=input_model_obj)
 
         # Validate output data
-        if not isinstance(output_model_obj, operator_output_model):
+        if not isinstance(output_obj, operator_output_model):
             raise InvalidOperatorOutputError(operator_name=self.__class__.__name__)
 
         # Push XCom
-        xcom_obj = self.format_xcom(xcom_obj=output_model_obj.dict())
+        xcom_obj = self.format_xcom(output_obj=output_obj)
         self.push_xcom(xcom_obj=xcom_obj)
 
 
+    @classmethod
+    def dry_run(
+        cls,
+        operator_input: dict, 
+        operator_input_model: pydantic.BaseModel,
+        operator_output_model: pydantic.BaseModel, 
+        operator_secrets_model: pydantic.BaseModel = None,
+        secrets_input: dict = None,
+    ):
+        # Instantiate 
+        input_model_obj = operator_input_model(**operator_input)
+        secrets_model_obj = operator_secrets_model(**secrets_input) if operator_secrets_model else None
+
+        class DryOperator(cls):
+            def __init__(self, secrets):
+                self.secrets = secrets
+                self.logger = get_configured_logger(f"{self.__class__.__name__ }-dry-run")
+
+        dry_instance = DryOperator(secrets=secrets_model_obj)
+
+        # Run operator function
+        return cls.operator_function(self=dry_instance, input_model=input_model_obj)
+
+
     @abc.abstractmethod
     def operator_function(self):
         """
