@@ -10,12 +10,6 @@ def create_operator(
 
     # Operator name as used by Airflow
     NewOperator.__name__ = metadata.get("name", "BaseOperator")
-
-    # Operator version: update it to auto update the deployed version
-    NewOperator.__version__ = metadata.get("version", "0.1.0")
-
-    # Dockerfile to build Image for this function
-    NewOperator.__dockerfile__ = metadata.get("dockerfile", "Dockerfile")
     
     # Operator custom function
     NewOperator.operator_function = operator_function
