@@ -41,7 +41,12 @@ class OperatorMetadata(BaseModel):
         example="ExampleOperator", 
         # regex=    # TODO - regex for *Operator  
     )
+    description: str = Field(
+        description="Description of this Operator", 
+        example="This Operator executes ABCDEFG function.", 
+    )
     dependency: Dependency = None
+    container_resources: Dict = None # TODO - add model for the container_resources dictionary
     tags: List[str] = None
     style: Dict = None # TODO - add model for the style dictionary
     input_schema: InputSchema = None
